using System;
using System.Text;
using System.Drawing;
using System.Threading;
using System.Collections;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace DuckEmulator.HabboHotel.Outside.Rooms.Movement
{
    struct PathFinderNode
    {
        public int     F;
        public int     G;
        public int     H;  // f = gone + heuristic
        public int     X;
        public int     Y;
        public int     PX; // Parent
        public int     PY;
    }

   
    enum PathFinderNodeType
    {
        Start   = 1,
        End     = 2,
        Open    = 4,
        Close   = 8,
        Current = 16,
        Path    = 32
    }

    enum HeuristicFormula
    {
        Manhattan           = 1,
        MaxDXDY             = 2,
        DiagonalShortCut    = 3,
        Euclidean           = 4,
        EuclideanNoSQR      = 5,
        Custom1             = 6
    }

    class PathFinder : IPathFinder
    {
        private bool[,] mGrid = null;
        private double[,] mHeightmap = null;
        private double mMaximumAscend = 1.5f;
        private double mMaximumDescend = 4;
        private bool mAccountHeights = true;
        private PriorityQueueB<PathFinderNode> mOpen = new PriorityQueueB<PathFinderNode>(new ComparePFNode());
        private List<PathFinderNode> mClose = new List<PathFinderNode>();
        private bool mStop = false;
        private bool mStopped = true;
        private HeuristicFormula mFormula = HeuristicFormula.MaxDXDY;
        private int mHEstimate = 2;
        private bool mTieBreaker = true;
        private int mSearchLimit = 25000;
        private double mCompletedTime = 0;

        private int MaxX = new int();
        private int MaxY = new int();

        public PathFinder(bool[,] grid, double[,] heightmap, int mX, int mY)
        {
            mGrid = grid;
            mHeightmap = heightmap;

            MaxX = mX;
            MaxY = mY;
        }

        public bool AccountHeights
        {
            get { return mAccountHeights; }
            set { mAccountHeights = value; }
        }

        public double MaxAscend
        {
            get { return mMaximumAscend; }
            set { mMaximumAscend = value; }
        }

        public double MaxDescend
        {
            get { return mMaximumDescend; }
            set { mMaximumDescend = value; }
        }

        public double[,] Heightmap
        {
            get { return mHeightmap; }
            set { mHeightmap = value; }
        }

        public bool Stopped
        {
            get { return mStopped; }
        }

        public HeuristicFormula Formula
        {
            get { return mFormula; }
            set { mFormula = value; }
        }

        public int HeuristicEstimate
        {
            get { return mHEstimate; }
            set { mHEstimate = value; }
        }

        public bool TieBreaker
        {
            get { return mTieBreaker; }
            set { mTieBreaker = value; }
        }

        public int SearchLimit
        {
            get { return mSearchLimit; }
            set { mSearchLimit = value; }
        }

        public double CompletedTime
        {
            get { return mCompletedTime; }
            set { mCompletedTime = value; }
        }

        public void FindPathStop()
        {
            mStop = true;
        }

        public List<PathFinderNode> FindPath(Point start, Point end)
        {
            lock (this)
            {
                PathFinderNode parentNode = new PathFinderNode();
                bool found = false;

                mStop = false;
                mStopped = false;
                mOpen.Clear();
                mClose = new List<PathFinderNode>();

                sbyte[,] direction = new sbyte[8, 2] { { 0, -1 }, { 1, 0 }, { 0, 1 }, { -1, 0 }, { -1, -1 }, { 1, -1 }, { 1, 1 }, { -1, 1 } };

                parentNode.G = 0;
                parentNode.H = mHEstimate;
                parentNode.F = parentNode.G + parentNode.H;
                parentNode.X = start.X;
                parentNode.Y = start.Y;
                parentNode.PX = parentNode.X;
                parentNode.PY = parentNode.Y;

                mOpen.Push(parentNode);

                while (mOpen.Count > 0 && !mStop)
                {
                    parentNode = mOpen.Pop();

                    if (parentNode.X == end.X && parentNode.Y == end.Y)
                    {
                        mClose.Add(parentNode);
                        found = true;
                        break;
                    }

                    if (mClose.Count > mSearchLimit)
                    {
                        mStopped = true;
                        return null;
                    }

                    //Lets calculate each successors
                    for (int i = 0; i < 8; i++)
                    {
                        PathFinderNode newNode;
                        newNode.X = parentNode.X + direction[i, 0];
                        newNode.Y = parentNode.Y + direction[i, 1];

                        if (newNode.X < 0 || newNode.Y < 0 || newNode.X >= MaxX || newNode.Y >= MaxY)
                        {
                            continue;
                        }

                        bool State = true; try { State = mGrid[newNode.X, newNode.Y]; }
                        catch { State = false; }

                        int newG = parentNode.G + ((State) ? 1 : 0);

                        if (newG == parentNode.G)
                        {
                            //Unbrekeable
                            continue;
                        }

                        //Take into account heights
                        if (mAccountHeights)
                        {
                            if (mMaximumDescend > -1)
                            {
                                if (mHeightmap[parentNode.X, parentNode.Y] - mMaximumDescend > mHeightmap[newNode.X, newNode.Y])
                                {
                                    continue;
                                }
                            }

                            if (mMaximumAscend > -1)
                            {
                                if (mHeightmap[parentNode.X, parentNode.Y] + mMaximumAscend < mHeightmap[newNode.X, newNode.Y])
                                {
                                    continue;
                                }
                            }
                        }

                        int foundInOpenIndex = -1;
                        for (int j = 0; j < mOpen.Count; j++)
                        {
                            if (mOpen[j].X == newNode.X && mOpen[j].Y == newNode.Y)
                            {
                                foundInOpenIndex = j;
                                break;
                            }
                        }
                        if (foundInOpenIndex != -1 && mOpen[foundInOpenIndex].G <= newG)
                            continue;

                        int foundInCloseIndex = -1;
                        for (int j = 0; j < mClose.Count; j++)
                        {
                            if (mClose[j].X == newNode.X && mClose[j].Y == newNode.Y)
                            {
                                foundInCloseIndex = j;
                                break;
                            }
                        }
                        if (foundInCloseIndex != -1 && mClose[foundInCloseIndex].G <= newG)
                            continue;

                        newNode.PX = parentNode.X;
                        newNode.PY = parentNode.Y;
                        newNode.G = newG;

                        switch (mFormula)
                        {
                            default:
                            case HeuristicFormula.Manhattan:
                                newNode.H = mHEstimate * (Math.Abs(newNode.X - end.X) + Math.Abs(newNode.Y - end.Y));
                                break;
                            case HeuristicFormula.MaxDXDY:
                                newNode.H = mHEstimate * (Math.Max(Math.Abs(newNode.X - end.X), Math.Abs(newNode.Y - end.Y)));
                                break;
                            case HeuristicFormula.DiagonalShortCut:
                                int h_diagonal = Math.Min(Math.Abs(newNode.X - end.X), Math.Abs(newNode.Y - end.Y));
                                int h_straight = (Math.Abs(newNode.X - end.X) + Math.Abs(newNode.Y - end.Y));
                                newNode.H = (mHEstimate * 2) * h_diagonal + mHEstimate * (h_straight - 2 * h_diagonal);
                                break;
                            case HeuristicFormula.Euclidean:
                                newNode.H = (int)(mHEstimate * Math.Sqrt(Math.Pow((newNode.X - end.X), 2) + Math.Pow((newNode.Y - end.Y), 2)));
                                break;
                            case HeuristicFormula.EuclideanNoSQR:
                                newNode.H = (int)(mHEstimate * (Math.Pow((newNode.X - end.X), 2) + Math.Pow((newNode.Y - end.Y), 2)));
                                break;
                            case HeuristicFormula.Custom1:
                                Point dxy = new Point(Math.Abs(end.X - newNode.X), Math.Abs(end.Y - newNode.Y));
                                int Orthogonal = Math.Abs(dxy.X - dxy.Y);
                                int Diagonal = Math.Abs(((dxy.X + dxy.Y) - Orthogonal) / 2);
                                newNode.H = mHEstimate * (Diagonal + Orthogonal + dxy.X + dxy.Y);
                                break;
                        }
                        if (mTieBreaker)
                        {
                            int dx1 = parentNode.X - end.X;
                            int dy1 = parentNode.Y - end.Y;
                            int dx2 = start.X - end.X;
                            int dy2 = start.Y - end.Y;
                            int cross = Math.Abs(dx1 * dy2 - dx2 * dy1);
                            newNode.H = (int)(newNode.H + cross * 0.001);
                        }
                        newNode.F = newNode.G + newNode.H;

                        mOpen.Push(newNode);
                    }

                    mClose.Add(parentNode);

                }

                if (found)
                {
                    PathFinderNode fNode = mClose[mClose.Count - 1];
                    for (int i = mClose.Count - 1; i >= 0; i--)
                    {
                        if (fNode.PX == mClose[i].X && fNode.PY == mClose[i].Y || i == mClose.Count - 1)
                        {
                            fNode = mClose[i];
                        }
                        else
                            mClose.RemoveAt(i);
                    }
                    mStopped = true;
                    return mClose;
                }
                mStopped = true;
                return null;
            }
        }

        internal class ComparePFNode : IComparer<PathFinderNode>
        {
            public int Compare(PathFinderNode x, PathFinderNode y)
            {
                if (x.F > y.F)
                    return 1;
                else if (x.F < y.F)
                    return -1;
                return 0;
            }
        }
    }
}
